@@ -1,14 +1,21 @@
-from pandas import DataFrame, Series
-from sklearn.base import BaseEstimator, ClassifierMixin
+from sklearn.base import BaseEstimator
 from sklearn.exceptions import NotFittedError
-from sklearn.externals import joblib
-from sklearn.feature_selection.base import SelectorMixin
+try:
+	from sklearn.compose import ColumnTransformer
+except ImportError:
+	class ColumnTransformer:
+		pass
+try:
+	from sklearn.feature_selection._base import SelectorMixin
+except ImportError:
+	from sklearn.feature_selection.base import SelectorMixin
 from sklearn.feature_selection import SelectFromModel
 from sklearn.pipeline import FeatureUnion, Pipeline
 from sklearn_pandas import DataFrameMapper
-from subprocess import CalledProcessError
+from subprocess import PIPE, Popen
 from zipfile import ZipFile
 
+import joblib
 import numpy
 import os
 import pandas
@@ -17,127 +24,66 @@ import platform
 import re
 import sklearn
 import sklearn_pandas
-import subprocess
 import tempfile
 
 from .metadata import __copyright__, __license__, __version__
+from .pipeline import PMMLPipeline
 
-class _Verification(object):
-
-	def __init__(self, active_values, target_values, precision, zeroThreshold):
-		if precision < 0:
-			raise ValueError("Precision cannot be negative")
-		if zeroThreshold < 0:
-			raise ValueError("Zero threshold cannot be negative")
-		self.active_values = active_values
-		self.target_values = target_values
-		self.precision = precision
-		self.zeroThreshold = zeroThreshold
-
-def _get_column_names(X):
-	if isinstance(X, DataFrame):
-		return X.columns.values
-	elif isinstance(X, Series):
-		return numpy.asarray(X.name)
-	else:
-		return None
-
-def _get_values(X):
-	if isinstance(X, DataFrame):
-		return X.values
-	elif isinstance(X, Series):
-		return X.values
-	else:
-		return X
-
-class PMMLPipeline(Pipeline):
-
-	def __init__(self, steps):
-		super(PMMLPipeline, self).__init__(steps = steps)
-
-	def __repr__(self):
-		class_name = self.__class__.__name__
-		return "%s(steps=[%s])" % (class_name, (",\n" + (1 + len(class_name) // 2) * " ").join(repr(step) for step in self.steps))
-
-	def _fit(self, X, y = None, **fit_params):
-		# Collect feature name(s)
-		active_fields = _get_column_names(X)
-		if active_fields is not None:
-			self.active_fields = active_fields
-		# Collect label name(s)
-		target_fields = _get_column_names(y)
-		if target_fields is not None:
-			self.target_fields = target_fields
-		return super(PMMLPipeline, self)._fit(X = X, y = y, **fit_params)
-
-	def verify(self, X, precision = 1e-13, zeroThreshold = 1e-13):
-		active_fields = _get_column_names(X)
-		if self.active_fields is None or active_fields is None:
-			raise ValueError("Cannot perform model validation with anonymous data")
-		if self.active_fields.tolist() != active_fields.tolist():
-			raise ValueError("The columns between training data {} and verification data {} do not match".format(self.active_fields, active_fields))
-		active_values = _get_values(X)
-		y = self.predict(X)
-		target_values = _get_values(y)
-		self.verification = _Verification(active_values, target_values, precision, zeroThreshold)
-		estimator = self._final_estimator
-		if isinstance(estimator, ClassifierMixin) and hasattr(estimator, "predict_proba"):
-			try:
-				y_proba = self.predict_proba(X)
-				self.verification.probability_values = _get_values(y_proba)
-			except AttributeError:
-				pass
+def _is_categorical(dtype):
+	if dtype == object or dtype == str or dtype == bool:
+		return True
+	elif hasattr(dtype, "name") and dtype.name == "category":
+		return True
+	return False
 
 class EstimatorProxy(BaseEstimator):
 
-	def __init__(self, estimator_, attr_names_ = ["feature_importances_"]):
-		self.estimator_ = estimator_
-		self.attr_names_ = attr_names_
+	def __init__(self, estimator, attr_names = ["feature_importances_"]):
+		self.estimator = estimator
+		self.attr_names = attr_names
 		try:
 			self._copy_attrs()
 		except NotFittedError:
 			pass
 
-	def __getattr__(self, name):
-		return getattr(self.estimator_, name)
-
 	def _copy_attrs(self):
-		for attr_name_ in self.attr_names_:
-			if hasattr(self.estimator_, attr_name_):
-				setattr(self, attr_name_, getattr(self.estimator_, attr_name_))
+		for attr_name in self.attr_names:
+			if hasattr(self.estimator, attr_name):
+				setattr(self, attr_name, getattr(self.estimator, attr_name))
 
 	def fit(self, X, y = None, **fit_params):
-		self.estimator_.fit(X, y, **fit_params)
+		self.estimator.fit(X, y, **fit_params)
 		self._copy_attrs()
 		return self
 
+	def predict(self, X, **predict_params):
+		return self.estimator.predict(X, **predict_params)
+
+	def predict_proba(self, X, **predict_proba_params):
+		return self.estimator.predict_proba(X, **predict_proba_params)
+
 class SelectorProxy(BaseEstimator):
 
-	def __init__(self, selector_):
-		self.selector_ = selector_
+	def __init__(self, selector):
+		self.selector = selector
 		try:
 			self._copy_attrs()
 		except NotFittedError:
 			pass
 
-	def __getattr__(self, name):
-		return getattr(self.selector_, name)
-
 	def _copy_attrs(self):
 		try:
-			setattr(self, "support_mask_", self.selector_._get_support_mask())
+			setattr(self, "support_mask_", self.selector._get_support_mask())
 		except ValueError:
 			pass
 
 	def fit(self, X, y = None, **fit_params):
-		self.selector_.fit(X, y, **fit_params)
+		self.selector.fit(X, y, **fit_params)
 		self._copy_attrs()
 		return self
 
-	def fit_transform(self, X, y = None, **fit_params):
-		Xt = self.selector_.fit_transform(X, y, **fit_params)
-		self._copy_attrs()
-		return Xt
+	def transform(self, X):
+		return self.selector.transform(X)
 
 def _get_steps(obj):
 	if isinstance(obj, Pipeline):
@@ -153,16 +99,16 @@ def _filter(obj):
 		if hasattr(obj, "built_features"):
 			if obj.built_features is not None:
 				obj.built_features = _filter_steps(obj.built_features)
+	elif isinstance(obj, ColumnTransformer):
+		obj.transformers = _filter_steps(obj.transformers)
+		obj.remainder = _filter(obj.remainder)
+		if hasattr(obj, "transformers_"):
+			obj.transformers_ = _filter_steps(obj.transformers_)
 	elif isinstance(obj, FeatureUnion):
 		obj.transformer_list = _filter_steps(obj.transformer_list)
 	elif isinstance(obj, Pipeline):
 		obj.steps = _filter_steps(obj.steps)
 	elif isinstance(obj, SelectorMixin):
-		if isinstance(obj, SelectFromModel):
-			if hasattr(obj, "estimator"):
-				setattr(obj, "estimator", EstimatorProxy(obj.estimator))
-			if hasattr(obj, "estimator_"):
-				setattr(obj, "estimator_", EstimatorProxy(obj.estimator_))
 		return SelectorProxy(obj)
 	elif isinstance(obj, list):
 		return [_filter(e) for e in obj]
@@ -194,6 +140,27 @@ def make_pmml_pipeline(obj, active_fields = None, target_fields = None):
 		pipeline.target_fields = numpy.asarray(target_fields)
 	return pipeline
 
+def _decode(data, encoding):
+	try:
+		return data.decode(encoding, errors = "ignore")
+	except ValueError:
+		return ""
+
+def _java_version(java_encoding):
+	try:
+		process = Popen(["java", "-version"], stdout = PIPE, stderr = PIPE, bufsize = 1)
+	except:
+		return None
+	output, error = process.communicate()
+	retcode = process.poll()
+	if retcode:
+		return None
+	match = re.match("^(.*)\sversion\s\"(.*)\"(|\s\d\d\d\d\-\d\d\-\d\d)$", _decode(error, java_encoding), re.MULTILINE)
+	if match:
+		return (match.group(1), match.group(2))
+	else:
+		return None
+
 def _package_classpath():
 	jars = []
 	resources = pkg_resources.resource_listdir("sklearn2pmml.resources", "")
@@ -224,16 +191,16 @@ def _dump(obj, prefix):
 		os.close(fd)
 	return path
 
-def sklearn2pmml(pipeline, pmml, user_classpath = [], with_repr = False, debug = False):
-	"""Converts fitted Scikit-Learn pipeline to PMML.
+def sklearn2pmml(pipeline, pmml, user_classpath = [], with_repr = False, debug = False, java_encoding = "UTF-8"):
+	"""Converts a fitted PMML pipeline object to PMML file.
 
 	Parameters:
 	----------
 	pipeline: PMMLPipeline
-		The pipeline.
+		The input PMML pipeline object.
 
 	pmml: string
-		The path to where the PMML document should be stored.
+		The output PMML file.
 
 	user_classpath: list of strings, optional
 		The paths to JAR files that provide custom Transformer, Selector and/or Estimator converter classes.
@@ -243,36 +210,63 @@ def sklearn2pmml(pipeline, pmml, user_classpath = [], with_repr = False, debug =
 		If true, insert the string representation of pipeline into the PMML document.
 
 	debug: boolean, optional
-		If true, print information about the conversion operation.
+		If true, print information about the conversion process.
+
+	java_encoding: string, optional
+		The character encoding to use for decoding Java output and error byte streams.
 
 	"""
 	if debug:
-		print("python: ", platform.python_version())
-		print("sklearn: ", sklearn.__version__)
-		print("sklearn.externals.joblib:", joblib.__version__)
-		print("pandas: ", pandas.__version__)
-		print("sklearn_pandas: ", sklearn_pandas.__version__)
-		print("sklearn2pmml: ", __version__)
+		java_version = _java_version(java_encoding)
+		if java_version is None:
+			java_version = ("java", "N/A")
+		print("python: {0}".format(platform.python_version()))
+		print("sklearn: {0}".format(sklearn.__version__))
+		print("sklearn2pmml: {0}".format(__version__))
+		print("joblib: {0}".format(joblib.__version__))
+		print("sklearn_pandas: {0}".format(sklearn_pandas.__version__))
+		print("pandas: {0}".format(pandas.__version__))
+		print("numpy: {0}".format(numpy.__version__))
+		print("{0}: {1}".format(java_version[0], java_version[1]))
 	if not isinstance(pipeline, PMMLPipeline):
-		raise TypeError("The pipeline object is not an instance of " + PMMLPipeline.__name__)
+		raise TypeError("The pipeline object is not an instance of " + PMMLPipeline.__name__ + ". Use the 'sklearn2pmml.make_pmml_pipeline(obj)' utility function to translate a regular Scikit-Learn estimator or pipeline to a PMML pipeline")
+	estimator = pipeline._final_estimator
 	cmd = ["java", "-cp", os.pathsep.join(_classpath(user_classpath)), "org.jpmml.sklearn.Main"]
 	dumps = []
 	try:
 		if with_repr:
 			pipeline.repr_ = repr(pipeline)
+		# if isinstance(estimator, H2OEstimator):
+		if hasattr(estimator, "download_mojo"):
+			estimator_mojo = estimator.download_mojo()
+			dumps.append(estimator_mojo)
+			estimator._mojo_path = estimator_mojo
 		pipeline_pkl = _dump(pipeline, "pipeline")
 		cmd.extend(["--pkl-pipeline-input", pipeline_pkl])
 		dumps.append(pipeline_pkl)
 		cmd.extend(["--pmml-output", pmml])
 		if debug:
-			print(" ".join(cmd))
+			print("Executing command:\n{0}".format(" ".join(cmd)))
 		try:
-			subprocess.check_call(cmd)
-		except CalledProcessError:
-			raise RuntimeError("The JPMML-SkLearn conversion application has failed. The Java process should have printed more information about the failure into its standard output and/or error streams")
+			process = Popen(cmd, stdout = PIPE, stderr = PIPE, bufsize = 1)
+		except OSError:
+			raise RuntimeError("Java is not installed, or the Java executable is not on system path")
+		output, error = process.communicate()
+		retcode = process.poll()
+		if debug or retcode:
+			if(len(output) > 0):
+				print("Standard output:\n{0}".format(_decode(output, java_encoding)))
+			else:
+				print("Standard output is empty")
+			if(len(error) > 0):
+				print("Standard error:\n{0}".format(_decode(error, java_encoding)))
+			else:
+				print("Standard error is empty")
+		if retcode:
+			raise RuntimeError("The JPMML-SkLearn conversion application has failed. The Java executable should have printed more information about the failure into its standard output and/or standard error streams")
 	finally:
 		if debug:
-			print("Preserved joblib dump file(s): ", " ".join(dumps))
+			print("Preserved joblib dump file(s): {0}".format(" ".join(dumps)))
 		else:
 			for dump in dumps:
 				os.remove(dump)
